from __future__ import unicode_literals

"""

.. module:: django-ewiz.decompiler
    :synopsis: django-ewiz database backend decompiler.

    django-ewiz is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    django-ewiz is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with django-ewiz. If not, see <http://www.gnu.org/licenses/>.

.. moduleauthor:: Alex Kavanaugh <kavanaugh.development@outlook.com>

"""

import sys
import re
from functools import wraps

from django.db.utils import DatabaseError

import requests

from .urlbuilders import Read


def safe_call(func):
    """Function wrapper for debugging - taken from Django-Nonrel/djangotoolbox."""

    @wraps(func)
    def _func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception, message:
            raise DatabaseError(unicode(str(message)) + unicode(str(sys.exc_info()[2])))

    return _func


class EwizDecompiler(object):
    """

    Ewiz results decompiler

    Send requests to the EnterpriseWizard database via the REST API and parses the response.

    """

    def __init__(self, model, settings_dict):
        self.model = model
        self.settings_dict = settings_dict

    @safe_call
    def decompile(self, url):
        """

        Requests tickets given a query url and parses each result.

        This method returns a list of field, value dictionaries. Each dictionary represents a ticket.

        """

        count, response_list = self.__request_multiple(url)
        query_list = []

        for response in response_list:
            query_list.append(self.__decompile(response))

        return query_list

    @safe_call
    def count(self, url):
        """

        Requests tickets given a query url and parses only the ticket count.

        This method returns the number of tickets the given query returned.

        """

        count, response_list = self.__request_multiple(url, count_only=True)

        return count

    def __request_multiple(self, url, count_only=False):
        """

        Parses a multiple ticket response into a list of responses (one for each ticket) and a count the number of tickets returned.

        Returns either the list of ticket responses or the count of tickets returned, depending on countOnly's value.

        """

        try:
            response = requests.get(url)
        except requests.exceptions.HTTPError, message:
            raise self.model.DoesNotExist(self.model._meta.object_name + ' matching query does not exist.\n\t' + str(message))

        pattern = re.compile(r"^EWREST_id_.* = '(?P<value>.*)';$", re.DOTALL)

        id_list = []
        for line in response.iter_lines(decode_unicode=True):
            try:
                id_list.append(pattern.match(line).group('value'))
            except:
                raise DatabaseError("Connection Error. The EnterpriseWizard database might be offline.")

        count = int(id_list[0])
        idList = id_list[1:]
        response_list = []

        # Return only the count before the heavy lifting if countOnly is True
        if count_only:
            return count, response_list

        for ticket_id in idList:
            response_url = Read(self.settings_dict, self.model._meta.db_table, ticket_id).build()
            response_list.append(self.__request_single(response_url))

        return count, response_list

    def __request_single(self, url):
        """Generates a response for a single ticket."""

        try:
            return requests.get(url)
        except requests.exceptions.HTTPError, message:
            raise self.model.DoesNotExist(self.model._meta.object_name + ' matching query does not exist.\n\t' + unicode(str(message)))

    def __decompile(self, response):
        """Parses a response into a field, value dictionary."""

        data_list = []
        for line in response.iter_lines(decode_unicode=True):
            data_list.append(line)

        pattern = re.compile(r"^EWREST_(?P<key>.*?)='(?P<value>.*)';$", re.DOTALL)

        data_dict = {}
        for data in data_list:
            match = pattern.match(data)
            data_dict[match.group('key')] = match.group('value')

        return data_dict



from __future__ import unicode_literals

"""

.. module:: django-ewiz.compiler
    :synopsis: django-ewiz database backend compilers.

    django-ewiz is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    django-ewiz is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with django-ewiz. If not, see <http://www.gnu.org/licenses/>.

.. moduleauthor:: Alex Kavanaugh <kavanaugh.development@outlook.com>

"""

import sys
import re
from functools import wraps

from django.db.utils import DatabaseError, IntegrityError
from django.db.models.sql.constants import SINGLE, MULTI
from django.db.models.sql import aggregates as sqlaggregates
from django.utils.datastructures import SortedDict

import requests
from djangotoolbox.db.basecompiler import (NonrelQuery, NonrelCompiler, NonrelInsertCompiler, NonrelUpdateCompiler, NonrelDeleteCompiler)

from .decompiler import EwizDecompiler
from .urlbuilders import Select, Update, Insert


def safe_call(func):
    """Function wrapper for debugging - taken from Django-Nonrel/djangotoolbox."""

    @wraps(func)
    def _func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception, message:
            raise DatabaseError(unicode(str(message)) + unicode(str(sys.exc_info()[2])))

    return _func


class EwizQuery(NonrelQuery):
    """

    The Ewiz class layer for nonrel queries.

    This class provides the framework for a query to the Ewiz REST interface.

    """

    # A dictionary of operators and their ewiz REST representations.
    operators = {
        'exact': lambda lookup_type, value: ("= BINARY", "'" + unicode(str(value)) + "'"),
        'iexact': lambda lookup_type, value: ("=", "'" + unicode(str(value)) + "'"),
        'contains': lambda lookup_type, value: ("LIKE BINARY", "'%" + unicode(str(value)) + "%'"),
        'icontains': lambda lookup_type, value: ("LIKE", "'%" + unicode(str(value)) + "%'"),
        'gt': lambda lookup_type, value: (">", "'" + unicode(str(value)) + "'"),
        'gte': lambda lookup_type, value: (">=", "'" + unicode(str(value)) + "'"),
        'lt': lambda lookup_type, value: ("<", "'" + unicode(str(value)) + "'"),
        'lte': lambda lookup_type, value: ("<=", "'" + unicode(str(value)) + "'"),
        'in': lambda lookup_type, value: ("IN", value),
        'startswith': lambda lookup_type, value: ("LIKE BINARY", "'" + unicode(str(value)) + "%'"),
        'istartswith': lambda lookup_type, value: ("LIKE", "'" + unicode(str(value)) + "%'"),
        'endswith': lambda lookup_type, value: ("LIKE BINARY", "'%" + unicode(str(value)) + "'"),
        'iendswith': lambda lookup_type, value: ("LIKE", "'%" + unicode(str(value)) + "'"),
        'range': lambda lookup_type, value: ("BETWEEN", unicode(str(value))),
        'isnull': lambda lookup_type, value: ("IS NULL", None),
        'regex': lambda lookup_type, value: ("REGEXP BINARY", "'" + unicode(str(value)) + "'"),
        'iregex': lambda lookup_type, value: ("REGEXP", "'" + unicode(str(value)) + "'"),
    }

    # A dictionary of operators and their negated ewiz REST representations.
    negated_operators = {
        'exact': lambda lookup_type, value: ("!= BINARY", "'" + unicode(str(value)) + "'"),
        'iexact': lambda lookup_type, value: ("!=", "'" + unicode(str(value)) + "'"),
        'contains': lambda lookup_type, value: ("NOT LIKE BINARY", "'%" + unicode(str(value)) + "%'"),
        'icontains': lambda lookup_type, value: ("NOT LIKE", "'%" + unicode(str(value)) + "%'"),
        'gt': lambda lookup_type, value: ("<", "'" + unicode(str(value)) + "'"),
        'gte': lambda lookup_type, value: ("<=", "'" + unicode(str(value)) + "'"),
        'lt': lambda lookup_type, value: (">", "'" + unicode(str(value)) + "'"),
        'lte': lambda lookup_type, value: (">=", "'" + unicode(str(value)) + "'"),
        'in': lambda lookup_type, value: ("NOT IN", value),
        'startswith': lambda lookup_type, value: ("NOT LIKE BINARY", "'" + unicode(str(value)) + "%'"),
        'istartswith': lambda lookup_type, value: ("NOT LIKE", "'" + unicode(str(value)) + "%'"),
        'endswith': lambda lookup_type, value: ("NOT LIKE BINARY", "'%" + unicode(str(value)) + "'"),
        'iendswith': lambda lookup_type, value: ("NOT LIKE", "'%" + unicode(str(value)) + "'"),
        'range': lambda lookup_type, value: ("NOT BETWEEN", unicode(str(value))),
        'isnull': lambda lookup_type, value: ("IS NOT NULL", None),
        'regex': lambda lookup_type, value: ("NOT REGEXP BINARY", "'" + unicode(str(value)) + "'"),
        'iregex': lambda lookup_type, value: ("NOT REGEXP", "'" + unicode(str(value)) + "'"),
    }

    def __init__(self, compiler, fields):
        super(EwizQuery, self).__init__(compiler, fields)
        self.compiled_query = {
            'table': None,
            'filters': [],
            'ordering': [],
            'limits': {
                'offset': '0',
                'limit': '18446744073709551615'  # Max limit as proposed by MySQL,
            },
        }

    def __debug(self):
        return ('DEBUG INFO:' +
            '\n\nRAW_QUERY: ' + str(self.query) +
            '\nCOMPILED_QUERY: ' + str(self.compiled_query) +
            '\nQUERY_URL: ' + str(Select(self.connection.settings_dict, self.query.model._meta.db_table, self.compiled_query).build())
        )

    @safe_call
    def fetch(self, low_mark=0, high_mark=None):
        """

        EwizQueryCompiler

        Gathers query parameters (filters, ordering, etc.) into a comiledQuery dictionary,
        builds a URL using those parameters (via URL Builders),
        and uses the decompiler to pull data using the REST API.

        This method returns an iterator (using for/yield) over the query results.

        """

        # Handle all records requests
        if not self.compiled_query["filters"]:
            self.compiled_query["filters"] = ["id LIKE '%'"]

        if high_mark is None:
            # Infinite fetching
            self.compiled_query["limits"]["offset"] = unicode(str(low_mark))
            self.compiled_query["limits"]["limit"] = '18446744073709551615'  # Max limit as proposed by MySQL
        elif high_mark > low_mark:
            # Range fetching
            self.compiled_query["limits"]["offset"] = unicode(str(low_mark))
            self.compiled_query["limits"]["limit"] = unicode(str(high_mark - low_mark))
        else:
            # Invalid range
            self.compiled_query["limits"]["offset"] = unicode(str(0))
            self.compiled_query["limits"]["limit"] = unicode(str(0))

        # Build the url
        url = Select(self.connection.settings_dict, self.query.model._meta.db_table, self.compiled_query).build()

        # Fetch and decompiler the results
        query_results = EwizDecompiler(self.query.model, self.connection.settings_dict).decompile(url)

        # Yield each result
        for result in query_results:
            yield result

    @safe_call
    def count(self, limit=None):
        """

        EwizQueryCounter

        Sends the query to Ewiz via the REST API, but only decompiles and returns the result count.

        This method returns the number of results a query will yield.

        """

        # Pass given limit to the compiled query
        if limit:
            self.compiled_query["limits"]["limit"] = str(unicode(limit))

        # Build the url
        url = Select(self.connection.settings_dict, self.query.model._meta.db_table, self.compiled_query).build()
        # Send the query, but only fetch and decompile the result count
        count = EwizDecompiler(self.query.model, self.connection.settings_dict).count(url)

        return count

    @safe_call
    def delete(self):
        raise NotImplementedError("Deleting EnterpriseWizard records is generally ill-advised. Please contact your EnterpriseWizard administrator for more information.")

    @safe_call
    def order_by(self, ordering):
        """

        EwizQueryOrderer

        Adds the ORDER BY parameter to the compiled query. Default: ORDER BY id ASC.

        NOTE:
        1) EnterpriseWizard's REST interface does not correctly represent ORDER BY queries. The Ewiz
           backend handles the query fine, but the frontend reorders it by ticketID. Since this causes
           limited results to be yielded incorrectly, this feature has been disabled until further notice.

        """

        # A True/False ordering designates default ordering.
        if type(ordering) is bool:
            if ordering:
                self.compiled_query["ordering"].append('id ASC')
            else:
                raise DatabaseError("The 'ORDER BY' statement is not currently supported by the EnterpriseWizard REST interface.")
                self.compiled_query["ordering"].append('id DESC')
        # A list of ordering tuples designates multiple ordering (non-default)
        else:
            for order in ordering:
                field = order[0]
                if order[1]:
                    direction = 'ASC'
                else:
                    direction = 'DESC'

                self.compiled_query["ordering"].append(field.name + ' ' + direction)

            raise DatabaseError("The 'ORDER BY' statement is not currently supported by the EnterpriseWizard REST interface.")

    @safe_call
    def add_filter(self, field, lookup_type, negated, value):
        """

        EwizQueryFilterer

        Adds a single constraint to be used in the WHERE clause of the compiled query.

        This method is called by the add_filters method of NonrelQuery

        """

        # Determine operator
        if negated:
            try:
                operator = self.negated_operators[lookup_type]
            except KeyError:
                raise DatabaseError("Lookup type %r can't be negated" % lookup_type)
        else:
            try:
                operator = self.operators[lookup_type]
            except KeyError:
                raise DatabaseError("Lookup type %r isn't supported" % lookup_type)

        # Handle lambda lookup types
        if callable(operator):
            operator, value = operator(lookup_type, value)

        self.compiled_query["filters"].append(field.column + ' ' + operator + ' ' + value)


class EwizCompiler(NonrelCompiler):
    """

    The Ewiz Query Compiler

    This handles SELECT queries and is the base class for INSERT, and UPDATE requests.

    """

    query_class = EwizQuery

    def execute_sql(self, result_type=MULTI):
        """

        Handles SQL-like aggregate queries. This method emulates COUNT
        by using the NonrelQuery.count method.

        """

        aggregates = self.query.aggregate_select.values()

        try:
            saveCheck = self.query.extra["a"] == ('1', [])
        except:
            pass

        # Simulate a count().
        if aggregates:
            assert len(aggregates) == 1
            aggregate = aggregates[0]
            assert isinstance(aggregate, sqlaggregates.Count)
            opts = self.query.get_meta()
            assert aggregate.col == '*' or \
                   aggregate.col == (opts.db_table, opts.pk.column)
            count = self.get_count()
            if result_type is SINGLE:
                return [count]
            elif result_type is MULTI:
                return [[count]]
        #
        # The save() method determines whether to UPDATE or INSERT based on whether or not primary_key query results exist (if a primary_key is not supplied, INSERT is chosen)
        # The following conditional section is a small hack that uses the count() method to count the number of results returned.
        #
        # False is returned if zero results are returned, True otherwise.
        #
        elif saveCheck:
            self.query.extra = SortedDict()
            count = self.get_count()
            if result_type is SINGLE:
                return (count != 0)
            elif result_type is MULTI:
                return (count != 0)

        raise NotImplementedError("The database backend only supports "
                                  "count() queries.")


class EwizInsertCompiler(NonrelInsertCompiler, EwizCompiler):
    """

    Ewiz Insert Compiler

    Creates new tickets in the urllib2Ewiz database via the REST API.

    """

    @safe_call
    def execute_sql(self, return_id=False):
        """

        Prepares field, value tuples for the compiler to use and,
        if requested, returns the primary key of the new ticket.

        """

        docs = []
        pk = self.query.get_meta().pk
        for obj in self.query.objs:
            doc = []
            for field in self.query.fields:
                value = field.get_db_prep_save(
                    getattr(obj, field.attname) if self.query.raw else field.pre_save(obj, obj._state.adding),
                    connection=self.connection
                )
                if not field.null and value is None and not field.primary_key:
                    raise IntegrityError("You can't set %s (a non-nullable "
                                         "field) to None!" % field.name)

                # Prepare value for database, note that query.values have
                # already passed through get_db_prep_save.
                value = self.ops.value_for_db(value, field)
                doc.append((field, value))

            docs.append(doc)

        if len(docs) > 1:
            raise DatabaseError('INSERT COMPILER: Docs length assumption was wrong. Contact Alex Kavanaugh with details./n/tDocs Length: ' + unicode(str(len(docs))))

        key = self.insert(docs[0], return_id=return_id)
        # Pass the key value through normal database deconversion.
        return self.ops.convert_values(self.ops.value_from_db(key, pk), pk)

    @safe_call
    def insert(self, values, return_id):
        """Builds and sends a query to create a new ticket in the Ewiz database."""

        # Build the url
        url = Insert(self.connection.settings_dict, self.query.model._meta.db_table, values).build()

        # Attempt the Insert
        try:
            response = requests.get(url)

            # Return the new ID
            if return_id:
                pattern = re.compile(r"^EWREST_id='(?P<value>.*)';$", re.DOTALL)

                line = next(response.iter_lines(decode_unicode=True))
                new_id = pattern.match(line).group('value')

                return int(new_id)

        except requests.exceptions.HTTPError, message:
            raise DatabaseError(self.query.model._meta.object_name + ' - An INSERT error has occurred. Please contact the development team with the following details:\n\t' + unicode(str(message)))


class EwizUpdateCompiler(NonrelUpdateCompiler):
    """

    Ewiz Insert Compiler

    Creates new tickets in the Ewiz database via the REST API.

    """

    @safe_call
    def update(self, values):
        """Builds and sends a query to update/change information in a ticket that currently exists in the Ewiz database."""

        # Build the url
        try:
            ticketID = self.query.where.children[0][-1]
        except:
            raise DatabaseError('UPDATE COMPILER: UPDATE ticketID assumptions were wrong. Contact Alex Kavanaugh with details.')
        url = Update(self.connection.settings_dict, self.query.model._meta.db_table, ticketID, values).build()

        # Attempt the Update
        try:
            requests.get(url)
        except requests.exceptions.HTTPError, message:
            raise DatabaseError(self.query.model._meta.object_name + ' - An UPDATE error has occurred. Please contact the development team with the following details:\n\t' + unicode(str(message)))


class EwizDeleteCompiler(NonrelDeleteCompiler):

    @safe_call
    def execute_sql(self, result_type=None):
        raise NotImplementedError("Deleting EnterpriseWizard records is generally ill-advised. Please contact your EnterpriseWizard administrator for more information.")

# Assign new compiler classes as default compilers
SQLCompiler = EwizCompiler
SQLInsertCompiler = EwizInsertCompiler
SQLUpdateCompiler = EwizUpdateCompiler
SQLDeleteCompiler = EwizDeleteCompiler



from distutils.core import setup

setup(
    name='django-ewiz',
    version='1.2.2',
    author='Alex Kavanaugh',
    author_email='kavanaugh.development@outlook.com',
    packages=['django_ewiz'],
    url='https://bitbucket.org/alex_kavanaugh/django-ewiz/',
    license='GNU LGPL (http://www.gnu.org/licenses/lgpl.html)',
    description="A non-relational Django database backend that utilizes EnterpriseWizard's REST interface.",
    long_description=open('README.rst').read(),
    install_requires=[
        "Django>=1.5",
        "djangotoolbox>=1.6.2",
        "requests>=2.2.0",
    ],
)



from __future__ import unicode_literals

"""

.. module:: django-ewiz.urlbuilders
    :synopsis: django-ewiz database backend URL builders.

    django-ewiz is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    django-ewiz is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with django-ewiz. If not, see <http://www.gnu.org/licenses/>.

.. moduleauthor:: Alex Kavanaugh <kavanaugh.development@outlook.com>

"""

import sys
import urllib
from functools import wraps

from django.db.utils import DatabaseError


def safe_call(func):
    """Function wrapper for debugging - taken from Django-Nonrel/djangotoolbox."""

    @wraps(func)
    def _func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception, message:
            raise DatabaseError(unicode(str(message)) + unicode(str(sys.exc_info()[2])))

    return _func


class Read(object):
    """

    Builds a READ url

    Sending a READ request returns a response containing fields and values for the provided ticket ID

    """

    def __init__(self, settings_dict, table, ticket_id):
        if settings_dict["PORT"] == "443":
            self.protocol = 'https://'
        else:
            self.protocol = 'http://'

        self.host = settings_dict["HOST"]
        self.knowledge_base = settings_dict["NAME"]
        self.login = settings_dict["USER"]
        self.password = settings_dict["PASSWORD"]
        self.language = 'en'
        self.table = table
        self.ticket_id = ticket_id

    @safe_call
    def build(self):
        return urllib.quote(self.protocol + self.host + 'EWRead?$KB=' + self.knowledge_base + '&$table=' + self.table + '&$login=' + self.login + '&$password=' + self.password + '&$lang=' + self.language + '&id=' + str(self.ticket_id), ":/?$&='")


class Select(object):
    """

    Builds a Select url

    Sending a SELECT request returns a response containing the number of tickets and a list of ticketIDs that adhere to the query constraints.

    """

    def __init__(self, settings_dict, table, compiled_query):
        if settings_dict["PORT"] == "443":
            self.protocol = 'https://'
        else:
            self.protocol = 'http://'

        self.host = settings_dict["HOST"]
        self.knowledge_base = settings_dict["NAME"]
        self.login = settings_dict["USER"]
        self.password = settings_dict["PASSWORD"]
        self.language = 'en'
        self.table = table
        self.compiled_query = compiled_query

    @safe_call
    def build(self):
        return urllib.quote(self.__build_select() + self.__build_where(), ":/?$&='")

    def __build_select(self):
        return self.protocol + self.host + 'EWSelect?$KB=' + self.knowledge_base + '&$table=' + self.table + '&$login=' + self.login + '&$password=' + self.password + '&$lang=' + self.language

    def __build_where(self):
        # Build filters string
        filters = ''
        for query_filter in self.compiled_query["filters"][:-1]:
            filters += query_filter + ' AND '
        else:
            filters += self.compiled_query["filters"][-1]

        # Build ordering string
        ordering = ' ORDER BY '
        for order in self.compiled_query["ordering"][:-1]:
            ordering += order + ', '
        else:
            ordering += self.compiled_query["ordering"][-1]

        # Build limits string
        limit = ' LIMIT ' + self.compiled_query["limits"]["limit"]
        offset = ' OFFSET ' + self.compiled_query["limits"]["offset"]

        return '&where=' + filters + ordering + limit + offset


class Insert(object):
    """

    Builds an Insert url

    Assuming there are no errors in the request, sending an INSERT request creates a ticket
    in the EnterpriseWizard database and returns the automatically generated ticket_id.

    """

    def __init__(self, settings_dict, table, data):
        if settings_dict["PORT"] == "443":
            self.protocol = 'https://'
        else:
            self.protocol = 'http://'

        self.host = settings_dict["HOST"]
        self.knowledge_base = settings_dict["NAME"]
        self.login = settings_dict["USER"]
        self.password = settings_dict["PASSWORD"]
        self.language = 'en'
        self.table = table
        self.data = data

    @safe_call
    def build(self):
        return urllib.quote(self.__build_insert() + self.__build_data() + '&time_spent=0:0:1:0', ":/?$&='")

    @safe_call
    def __build_insert(self):
        return self.protocol + self.host + 'EWCreate?$KB=' + self.knowledge_base + '&$table=' + self.table + '&$login=' + self.login + '&$password=' + self.password + '&$lang=' + self.language

    @safe_call
    def __build_data(self):
        data_string = ''
        for field, value in self.data:
            # Only insert if the field is editable and the field has a value or is allowed to be blank
            if (value or field.blank):  # field.editable and
                data_string += '&' + field.column + '=' + field.help_text + unicode(str(value))

        return data_string


class Update(object):
    """

    Builds an Update url

    Assuming there are no errors in the request, sending an UPDATE request updates/changes data
    in a ticket in the EnterpriseWizard database. Nothing is returned.

    """

    def __init__(self, settings_dict, table, ticket_id, data):
        if settings_dict["PORT"] == "443":
            self.protocol = 'https://'
        else:
            self.protocol = 'http://'

        self.host = settings_dict["HOST"]
        self.knowledge_base = settings_dict["NAME"]
        self.login = settings_dict["USER"]
        self.password = settings_dict["PASSWORD"]
        self.language = 'en'
        self.table = table
        self.ticket_id = unicode(str(ticket_id))
        self.data = data

    @safe_call
    def build(self):
        return urllib.quote(self.__build_update() + self.__build_data() + '&time_spent=0:0:1:0', ":/?$&='")

    @safe_call
    def __build_update(self):
        return self.protocol + self.host + 'EWUpdate?$KB=' + self.knowledge_base + '&$table=' + self.table + '&$login=' + self.login + '&$password=' + self.password + '&$lang=' + self.language

    @safe_call
    def __build_data(self):
        data_string = '&id=' + self.ticket_id
        for field, value in self.data:
            # Only update if the field is editable and the field has a value or is allowed to be blank
            if field.editable and (value or field.blank):
                data_string += '&' + field.column + '=' + field.help_text + unicode(str(value))

        return data_string


class Attach(object):
    """

    Builds an ATTACH url

    The ATTACH request must be sent via HTTP PUT. The number of attached files is returned upon upload success.

    """

    def __init__(self, settings_dict, table, ticket_id, field_name, file_name):
        if settings_dict["PORT"] == "443":
            self.protocol = 'https://'
        else:
            self.protocol = 'http://'

        self.host = settings_dict["HOST"]
        self.knowledge_base = settings_dict["NAME"]
        self.login = settings_dict["USER"]
        self.password = settings_dict["PASSWORD"]
        self.language = 'en'
        self.table = table
        self.ticket_id = ticket_id
        self.field_name = field_name
        self.file_name = file_name

    @safe_call
    def build(self):
        return urllib.quote(self.protocol + self.host + 'EWAttach?$KB=' + self.knowledge_base + '&$table=' + self.table + '&$login=' + self.login + '&$password=' + self.password + '&$lang=' + self.language + '&id=' + str(self.ticket_id) + '&field=' + str(self.field_name) + '&fileName=' + str(self.file_name), ":/?$&='")
